// ttyPort.cpp

/*************************************************************************
* Copyright (c) 2002 The University of Chicago, as Operator of Argonne
* National Laboratory, and the Regents of the University of California, as
* Operator of Los Alamos National Laboratory. MPFserial is distributed
* subject to a Software License Agreement found in file LICENSE that
* is included with this distribution.
*************************************************************************/
/* Author: Marty Kraimer
   Date: 07JAN2003
*/

#include <stdlib.h>
#include <stddef.h>
#include <string.h>
#include <stdio.h>
#include <errno.h>
#include <fcntl.h>
#include <termios.h>
#include <unistd.h>

#include <iocsh.h>
#include <epicsExport.h>
#include <epicsEvent.h>
#include <epicsThread.h>
#include <epicsTime.h>
#include <epicsAssert.h>
#include <iocsh.h>
#include <mpfType.h>
#include "SerialPort.h"

extern "C" { 
int initTtyPort(const char *portName, const char *fileName,
    int baud,const char *parity, int stopBits,int bitsPerChar,
    const char *flowType,int bufsize);
}

class TtyPort : public SerialPort{
public:
    TtyPort(const char *portName, const char *fileName,
        int baud, int stopBits, int bitsPerChar,char parity, char flowControl,
        int bufsize);
    bool open();
    virtual serialStatus write(
	unsigned char *data, int nbytes, double timeout);
    virtual serialStatus read(double timeout);
    virtual void flush();
    virtual bool config( int baud, int stopBits, int bitsPerChar,
	char parity, char flowControl);
    bool config();
    virtual void getConfig( int *baud, int *stopBits, int *bitsPerChar,
	char *parity, char *flowControl);
private:
    const char *portName;
    const char *fileName;
    int baud;
    int stopBits;
    int bitsPerChar;
    char parity;
    char flowControl;
    int fd;
    int bufsize;
    char *buffer;
    int bufcnt;  /* Number of characters in buffer. */
};

int initTtyPort(const char *port, const char *file,
    int baud,const char *parity,
    int stopBits,int bitsPerChar,
    const char *flowControl,int bufsize)
{
    char *portName,*fileName;
    if(bufsize<=1) {
        printf("bufsize %d is illegal. It is set to 100.\n",bufsize);
        bufsize = 100;
    }
    errno = 0;
    int fd = ::open(file,O_RDWR|O_NOCTTY);
    if(fd<0){
        printf("initTtyPort: open failed %s\n",strerror(errno));
        return(0);
    }
    /* Use lazy open in read/write.
     * If server issues all read/writes from a single thread then
     * all file access will be done by a single thread. */
    close(fd);
    portName = (char *)calloc(strlen(port)+1,sizeof(char));
    strcpy(portName,port);
    fileName = (char *)calloc(strlen(file)+1,sizeof(char));
    strcpy(fileName,file);
    TtyPort *pTtyPort = new TtyPort(portName,fileName,
	    baud,stopBits,bitsPerChar,parity[0],flowControl[0],bufsize);
    pTtyPort->registerName(portName);
    return(0);
}

TtyPort::TtyPort(const char *port, const char *file,
    int b, int sb, int bpc, char p, char fc, int bsize)
: portName(port), fileName(file), baud(b),
    stopBits(sb), bitsPerChar(bpc), parity(p), flowControl(fc),
    fd(-1), bufsize(bsize), bufcnt(0)
{
    buffer = (char *)calloc(bufsize,sizeof(char));
}

bool TtyPort::open()
{
    if(fd>=0) return(true);
    errno = 0;
    fd = ::open(fileName,O_RDWR|O_NOCTTY);
    if(fd<0){
        printf("TtyPort:open failed %s\n",strerror(errno));
        return(false);
    }
    if(!config()) {
        printf("TtyPort::open %s config failed\n",fileName);
        return(false);
    }
    return(true);
}

bool TtyPort::config( int b, int sb, int bpc, char p, char fc)
{
    baud = b; stopBits = sb; bitsPerChar = bpc; parity = p; flowControl = fc;
    if(fd<0 ) {if(!open()) return(false);}
    return(config());
}

bool TtyPort::config()
{
    struct termios tio;

    memset(&tio,0,sizeof(struct termios));
    tcgetattr(fd,&tio);
    tio.c_cflag |= (CLOCAL | CREAD);
    switch(baud) {
    case 1200: cfsetispeed(&tio,B1200); cfsetospeed(&tio,B1200); break;
    case 2400: cfsetispeed(&tio,B2400); cfsetospeed(&tio,B2400); break;
    case 4800: cfsetispeed(&tio,B4800); cfsetospeed(&tio,B4800); break;
    case 9600: cfsetispeed(&tio,B9600); cfsetospeed(&tio,B9600); break;
    case 19200: cfsetispeed(&tio,B19200); cfsetospeed(&tio,B19200); break;
    case 38400: cfsetispeed(&tio,B38400); cfsetospeed(&tio,B38400); break;
    case 0: break;
    default: printf("Illegal baud rate\n"); return(false);
    }
    tio.c_cflag &= ~CSTOPB;
    switch(stopBits) {
    case 1: break;
    case 2: tio.c_cflag |= CSTOPB; break;
    default: printf("Illegal number of stopBits\n"); return(false);
    }
    tio.c_cflag &= ~CSIZE;
    switch(bitsPerChar) {
    case 5: tio.c_cflag |= CS5; break;
    case 6: tio.c_cflag |= CS6; break;
    case 7: tio.c_cflag |= CS7; break;
    case 8: tio.c_cflag |= CS8; break;
    default: printf("Illegal number of bits\n"); return(false);
    }
    tio.c_cflag &= ~(PARENB|PARODD);
    switch(parity) {
    case 'E': tio.c_cflag |= (PARENB) ; break;
    case 'O': tio.c_cflag |= (PARENB|PARODD) ; break;
    case 'N': break;
    default: printf("Illegal parity");
    }
#ifdef CRTSCTS
    tio.c_cflag &= ~CRTSCTS;
    switch(flowControl) {
    case 'H': tio.c_cflag |= CRTSCTS ; break;
    case 'N': break;
    default: printf("Illegal flowControl");
    }
#endif
    tio.c_iflag = 0;
    tio.c_oflag = 0;
    tio.c_lflag = 0;
    /* read timout code requires VMIN=0 and VTIME=1 */
    tio.c_cc[VTIME] = 1;
    tio.c_cc[VMIN] = 0;
    tcsetattr(fd,TCSANOW,&tio);
    return(true);
}

void TtyPort::getConfig(int *gbaud, int *gstopBits, int *gbitsPerChar,
    char *gparity, char *gflowControl)
{
    if(fd<0 ) {if(!open()) return;}
    *gbaud = 0; *gstopBits=0; *gbitsPerChar=0; *gparity = 0; *gflowControl = 0;
    struct termios tio;
    memset(&tio,0,sizeof(struct termios));
    tcgetattr(fd,&tio);
    speed_t rate = cfgetispeed(&tio);
    switch(rate) {
        case B1200: *gbaud = 1200; break;
        case B2400: *gbaud = 2400; break;
        case B4800: *gbaud = 4800; break;
        case B9600: *gbaud = 9600; break;
        case B19200: *gbaud = 19200; break;
        case B38400: *gbaud = 38400; break;
        default: printf("TtyPort::getConfig illegal baudRate\n");
    }
    *gstopBits = (tio.c_cflag&CSTOPB) ? 2 : 1;
    tcflag_t bPerChar = tio.c_cflag&CSIZE;
    switch(bPerChar) {
        case CS5: *gbitsPerChar = 5; break;
        case CS6: *gbitsPerChar = 6; break;
        case CS7: *gbitsPerChar = 7; break;
        case CS8: *gbitsPerChar = 8; break;
        default: printf("TtyPort::getConfig Illegal bitsPerChar\n");
    }
    tcflag_t par = tio.c_cflag&(PARENB|PARODD);
    if(par==(PARENB|PARODD)) *gparity = 'O';
    else if(par==(PARENB)) *gparity = 'E';
    else *gparity = 'N';
#ifdef CRTSCTS
    if((tio.c_cflag&CRTSCTS) == CRTSCTS)
        *gflowControl = 'H';
#endif
}

serialStatus TtyPort::write(
    unsigned char *data, int nbytes, double )
{
    if(fd<0 ) {if(!open()) return(serialFailure);}
    errno = 0;
    int nout = ::write(fd,data,nbytes);
    if(nout!=nbytes) {
        printf("%s TtyPort::write nbytes %d but only %d written\n",
            fileName,nbytes,nout);
        return(serialFailure);
    }
    if(peekHandler) {
        for(int ind=0; ind<nout; ind++) {
            (*peekHandler)(peekHandlerParm,false,data[ind]);
        }
    }
    return(serialSuccess);
}

serialStatus TtyPort::read(double timeOut)
{
    byteHandlerRC rc = byteHandlerError;
    epicsTimeStamp start,now;

    if(fd<0 ) {if(!open()) return(serialFailure);}
    assert(byteHandler);
    epicsTimeGetCurrent(&start);
    errno = 0;
    while(1) {
        int ind, nin;

        errno = 0;
        if(bufcnt==0)  bufcnt = ::read(fd,buffer,bufsize);
        nin = bufcnt;
        if(nin<0) {
            printf("%s TtyPort::read failure %s\n",fileName,strerror(errno));
            return(serialFailure);
        }
        if(nin==0) {
            epicsTimeGetCurrent(&now);
            if(epicsTimeDiffInSeconds(&now,&start)>timeOut) {
                return(serialTimeout);
            }
            continue; /*Just try again. Note that VTIME MUST be > 0 */
        }
        for(ind=0; ind<nin; ind++) {
            bufcnt--;
            if(peekHandler) (*peekHandler)
                (peekHandlerParm,true,(unsigned char)buffer[ind]);
            rc = (*byteHandler)(byteHandlerParm,(unsigned char)buffer[ind]);
            if(rc!=byteHandlerOK) {
                serialStatus ioStatus = (rc==byteHandlerEndRead)
                    ? serialSuccess : serialFailure;
                if(bufcnt>0) memmove(buffer, &buffer[ind+1], bufcnt);
                return(ioStatus);
            }
        }
    }
}

void TtyPort::flush()
{
    bufcnt = 0;
    if(fd<0 ) {if(!open()) return;}
    tcflush(fd,TCIOFLUSH);
}

extern "C" {
static const iocshArg initTtyPortArg0 = { "portName",iocshArgString};
static const iocshArg initTtyPortArg1 = { "fileName",iocshArgString};
static const iocshArg initTtyPortArg2 = { "baud",iocshArgInt};
static const iocshArg initTtyPortArg3 = { "parity",iocshArgString};
static const iocshArg initTtyPortArg4 = { "stopBits",iocshArgInt};
static const iocshArg initTtyPortArg5 = { "bitsPerChar",iocshArgInt};
static const iocshArg initTtyPortArg6 = { "flowType",iocshArgString};
static const iocshArg initTtyPortArg7 = { "bufsize",iocshArgInt};
static const iocshArg * const initTtyPortArgs[8] = {
    &initTtyPortArg0,&initTtyPortArg1,&initTtyPortArg2,
    &initTtyPortArg3,&initTtyPortArg4,&initTtyPortArg5,
    &initTtyPortArg6,&initTtyPortArg7};
static const iocshFuncDef initTtyPortFuncDef = {
    "initTtyPort",8,initTtyPortArgs};
static void initTtyPortCallFunc(const iocshArgBuf *args)
{
    initTtyPort(
        args[0].sval,args[1].sval,args[2].ival,
        args[3].sval,args[4].ival,args[5].ival,
        args[6].sval,args[7].ival);
}

static void ttyPortRegister(void)
{
    iocshRegister(&initTtyPortFuncDef,initTtyPortCallFunc);
}
epicsExportRegistrar(ttyPortRegister);
} // end extern "C"
